﻿#region Namespaces
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using Autodesk.Revit.DB;
#endregion // Namespaces

namespace ObjExport
{
  class ObjExporter : IJtFaceEmitter
  {
    bool _add_color = true;

    #region VertexLookupXyz
    /// <summary>
    /// A vertex lookup class to eliminate 
    /// duplicate vertex definitions.
    /// </summary>
    class VertexLookupXyz : Dictionary<XYZ, int>
    {
      #region XyzEqualityComparer
      /// <summary>
      /// Define equality for Revit XYZ points.
      /// Very rough tolerance, as used by Revit itself.
      /// </summary>
      class XyzEqualityComparer : IEqualityComparer<XYZ>
      {
        const double _sixteenthInchInFeet
          = 1.0 / ( 16.0 * 12.0 );

        public bool Equals( XYZ p, XYZ q )
        {
          return p.IsAlmostEqualTo( q,
            _sixteenthInchInFeet );
        }

        public int GetHashCode( XYZ p )
        {
          return Util.PointString( p ).GetHashCode();
        }
      }
      #endregion // XyzEqualityComparer

      public VertexLookupXyz()
        : base( new XyzEqualityComparer() )
      {
      }

      /// <summary>
      /// Return the index of the given vertex,
      /// adding a new entry if required.
      /// </summary>
      public int AddVertex( XYZ p )
      {
        return ContainsKey( p )
          ? this[p]
          : this[p] = Count;
      }
    }
    #endregion // VertexLookupXyz

    #region VertexLookupInt
    /// <summary>
    /// An integer-based 3D point class.
    /// </summary>
    class PointInt : IComparable<PointInt>
    {
      public int X { get; set; }
      public int Y { get; set; }
      public int Z { get; set; }

      //public PointInt( int x, int y, int z )
      //{
      //  X = x;
      //  Y = y;
      //  Z = z;
      //}

      const double _feet_to_mm = 25.4 * 12;

      static int ConvertFeetToMillimetres( double d )
      {
        return (int) ( _feet_to_mm * d + 0.5 );
      }

      public PointInt( XYZ p )
      {
        X = ConvertFeetToMillimetres( p.X );
        Y = ConvertFeetToMillimetres( p.Y );
        Z = ConvertFeetToMillimetres( p.Z );
      }

      public int CompareTo( PointInt a )
      {
        int d = X - a.X;

        if( 0 == d )
        {
          d = Y - a.Y;

          if( 0 == d )
          {
            d = Z - a.Z;
          }
        }
        return d;
      }
    }

    /// <summary>
    /// A vertex lookup class to eliminate 
    /// duplicate vertex definitions.
    /// </summary>
    class VertexLookupInt : Dictionary<PointInt, int>
    {
      #region PointIntEqualityComparer
      /// <summary>
      /// Define equality for integer-based PointInt.
      /// </summary>
      class PointIntEqualityComparer : IEqualityComparer<PointInt>
      {
        public bool Equals( PointInt p, PointInt q )
        {
          return 0 == p.CompareTo( q );
        }

        public int GetHashCode( PointInt p )
        {
          return (p.X.ToString() 
            + "," + p.Y.ToString() 
            + "," + p.Z.ToString())
            .GetHashCode();
        }
      }
      #endregion // PointIntEqualityComparer

      public VertexLookupInt()
        : base( new PointIntEqualityComparer() )
      {
      }

      /// <summary>
      /// Return the index of the given vertex,
      /// adding a new entry if required.
      /// </summary>
      public int AddVertex( PointInt p )
      {
        return ContainsKey( p )
          ? this[p]
          : this[p] = Count;
      }
    }
    #endregion // VertexLookupInt

    #region ColorLookup
    /// <summary>
    /// A colour lookup class to eliminate 
    /// duplicate material definitions.
    /// </summary>
    class ColorLookup : Dictionary<Color, int>
    {
      #region ColorEqualityComparer
      /// <summary>
      /// Define equality for colours.
      /// </summary>
      class ColorEqualityComparer : IEqualityComparer<Color>
      {
        public static int Compare( Color p, Color q )
        {
          if( null == p )
          {
            return null == q ? 0 : -1;
          }
          if( null == q )
          {
            return 1;
          }
          int d = p.Red - q.Red;

          if( 0 == d )
          {
            d = p.Blue - q.Blue;

            if( 0 == d )
            {
              d = p.Green - q.Green;
            }
          }
          return d;
        }

        public static bool EqualsJ( Color p, Color q )
        {
          return 0 == Compare( p, q );
        }

        public bool Equals( Color p, Color q )
        {
          return 0 == Compare( p, q );
        }

        public int GetHashCode( Color p )
        {
          return ( p.Red.ToString()
            + "," + p.Blue.ToString()
            + "," + p.Green.ToString() )
            .GetHashCode();
        }
      }
      #endregion // ColorEqualityComparer

      Color _current;

      public ColorLookup()
        : base( new ColorEqualityComparer() )
      {
        _current = null;
      }

      /// <summary>
      /// Add a new entry for the given colour,
      /// if needed. Return true if the given 
      /// colour differs from the current colour,
      /// and update the current colour.
      /// </summary>
      public bool AddColor( Color color )
      {
        if( !ContainsKey( color ) )
        {
          this[color] = Count;
        }

        bool rc = !ColorEqualityComparer.EqualsJ( 
          _current, color );

        _current = color;

        return rc;
      }
    }
    #endregion // ColorLookup

    //VertexLookupXyz _vertices;
    VertexLookupInt _vertices;
    ColorLookup _colors;

    /// <summary>
    /// List of triangles, defined as 
    /// triples of vertex indices.
    /// Colours are also stored in the 
    /// list, using a negative first index,
    /// with the three rgb values stored in the 
    /// following one, followed by a zero value
    /// to keep the values in triples.
    /// </summary>
    List<int> _triangles;

    /// <summary>
    /// Keep track of the number of faces processed.
    /// </summary>
    int _faceCount;

    /// <summary>
    /// Keep track of the number of triangles processed.
    /// Originally, we just returned _triangles.Count
    /// divided by 3, but that no longer works now that 
    /// colours may be stored as well.
    /// </summary>
    int _triangleCount;

    public ObjExporter()
    {
      _faceCount = 0;
      _triangleCount = 0;
      _vertices = new VertexLookupInt();
      _triangles = new List<int>();

      if( _add_color )
      {
        _colors = new ColorLookup();
      }
    }

    /// <summary>
    /// Set a colour for the following faces.
    /// </summary>
    void StoreColor( Color color )
    {
      _triangles.Add( -1 ); // color marker
      _triangles.Add( Util.ColorToInt( color ) );
      _triangles.Add( 0 ); // multiple of three
    }

    /// <summary>
    /// Add the vertices of the given triangle to our
    /// vertex lookup dictionary and emit a triangle.
    /// </summary>
    void StoreTriangle( MeshTriangle triangle )
    {
      for( int i = 0; i < 3; ++i )
      {
        XYZ p = triangle.get_Vertex( i );
        PointInt q = new PointInt( p );
        _triangles.Add( _vertices.AddVertex( q ) );
      }
    }

    /// <summary>
    /// Emit a Revit geometry Face object and 
    /// return the number of resulting triangles.
    /// </summary>
    public int EmitFace( Face face, Color color )
    {
      ++_faceCount;

      if( _add_color && _colors.AddColor( color ) )
      {
        StoreColor( color );
      }

      Mesh mesh = face.Triangulate();

      int n = mesh.NumTriangles;

      Debug.Print( " {0} mesh triangles", n );

      for( int i = 0; i < n; ++i )
      {
        ++_triangleCount;

        MeshTriangle t = mesh.get_Triangle( i );

        StoreTriangle( t );
      }
      return n;
    }

    public int GetFaceCount()
    {
      return _faceCount;
    }

    /// <summary>
    /// Return the number of triangles processed.
    /// </summary>
    public int GetTriangleCount()
    {
      // Originally, we just returned _triangles.Count
      // divided by 3, but that no longer works now 
      // that colours may be stored as well.

      if( !_add_color )
      {
        int n = _triangles.Count;

        Debug.Assert( 0 == n % 3,
          "expected a multiple of 3" );

        Debug.Assert( _triangleCount.Equals( n / 3 ),
          "expected equal triangle count" );
      }
      return _triangleCount;
    }

    /// <summary>
    /// Return the number of uhnique vertices.
    /// </summary>
    public int GetVertexCount()
    {
      return _vertices.Count;
    }

    #region ExportTo: output the OBJ file
    /// <summary>
    /// Write a new colour definition to the 
    /// material library.
    /// </summary>
    static void EmitColor(
      StreamWriter s,
      Color color )
    {
      s.WriteLine( "newmtl {0}\r\n"
        + "Ka {1} {2} {3}\r\n"
        + "Kd {1} {2} {3}",
        Util.ColorString( color ),
        color.Red / 256.0, 
        color.Green / 256.0,
        color.Blue / 256.0 );
    }

    /// <summary>
    /// Obsolete: emit an XYZ vertex.
    /// </summary>
    static void EmitVertex( StreamWriter s, XYZ p )
    {
      s.WriteLine( "v {0} {1} {2}",
        Util.RealString( p.X ),
        Util.RealString( p.Y ),
        Util.RealString( p.Z ) );
    }

    /// <summary>
    /// Emit a vertex to OBJ. The first vertex listed 
    /// in the file has index 1, and subsequent ones
    /// are numbered sequentially.
    /// </summary>
    static void EmitVertex(
      StreamWriter s,
      PointInt p )
    {
      s.WriteLine( "v {0} {1} {2}", p.X, p.Y, p.Z );
    }

    /// <summary>
    /// Set colour for subsequent faces, referring
    /// to the named materials in the material 
    /// library.
    /// </summary>
    static void SetColor(
      StreamWriter s,
      int rgb )
    {
      s.WriteLine( "usemtl {0}",
        Util.ColorString( Util.IntToColor( rgb ) ) );
    }

    /// <summary>
    /// Emit an OBJ triangular face.
    /// </summary>
    static void EmitFacet(
      StreamWriter s,
      int i,
      int j,
      int k )
    {
      s.WriteLine( "f {0} {1} {2}",
        i + 1, j + 1, k + 1 );
    }

    public void ExportTo( string path )
    {
      string material_library_path = null;

      if( _add_color )
      {
        material_library_path = Path.ChangeExtension(
          path, "mtl" );

        using( StreamWriter s = new StreamWriter(
          material_library_path ) )
        {
          foreach( Color key in _colors.Keys )
          {
            EmitColor( s, key );
          }
        }
      }

      using( StreamWriter s = new StreamWriter( path ) )
      {
        if( _add_color )
        {
          s.WriteLine( "mtllib {0}",
            Path.GetFileName(
              material_library_path ) );
        }

        foreach( PointInt key in _vertices.Keys )
        {
          EmitVertex( s, key );
        }

        int i = 0;
        int n = _triangles.Count;

        while( i < n )
        {
          int i1 = _triangles[i++];
          int i2 = _triangles[i++];
          int i3 = _triangles[i++];

          if( -1 == i1 )
          {
            SetColor( s, i2 );
          }
          else
          {
            EmitFacet( s, i1, i2, i3 );
          }
        }
      }
    }
    #endregion // ExportTo: output the OBJ file
  }
}
